/*
 * B_H_DiceTower.c
 *
 * Created: 12/15/2019 8:21:23 PM
 * Author : johna
 *
 * Copyright (C) 2020 by John Anderson <racerxr650r@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any 
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN 
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */ 

#include <avr/io.h>
#include <stdlib.h>
#include <util/delay_basic.h>

// Program constants **********************************************************
#define BLANK_OUTPUT				00
#define DISPLAY_BLANK_TIME			100
#define DISPLAY_CONFIG_TIME			500

// Busy Delay function ********************************************************
#define F_CPU_KHZ 1000  // CPU clock speed in kHz
#define LOOPS_PER_MS (uint16_t)(F_CPU_KHZ/4)

// Program data types *********************************************************
typedef enum {State_Init, State_Idle, State_Update_Count, State_Display_Count, State_Update_Dice, State_Display_Dice, State_Roll, State_Roll_Result}State;

// Busy wait delay function ***************************************************
void delay_ms(uint16_t ms)
{
	while(ms--)
		_delay_loop_2(LOOPS_PER_MS);
}

// Setup the IO pins **********************************************************
void initIO()
{
	// Set Port B pins to input
	DDRB = 0b00000000;
	// Enable pull ups on port B
	PORTB = 0b11111111;
	
	// Set Port C pins to input
	DDRC = 0b00000000;
	// Enable pull ups on Port C
	PORTC = 0b11111111;

	// Set Port D pins to output
	DDRD = 0xff;
	PCICR = 0x00;
	PCMSK2 = 0x00;
	EIMSK = 0x00;
	
}

// Output a two digit decimal number on the Nixie Tube display ****************
void output(int out)
{
	int outputByte = out%10;
	out /= 10;
	outputByte |= (out%10)<<4;
	
	PORTD = outputByte;
}

// Get the dice count from the thumbwheel input *******************************
int getCount()
{
	uint8_t rawCount = ~PINB;
	
	// Bad wiring on both the thumbwheel and the board
	int count = (rawCount&0b00000011) | ((rawCount&0b00010000)>>2) | ((rawCount&0b00000100)<<1);
	count += (((rawCount&0b00100000)>>5) | ((rawCount&0b00001000)>>2) | ((rawCount&0b11000000)>>4)) * 10;
	
	return(count);
}

// Get the dice type **********************************************************
int getDice()
{
	uint8_t rawDice = (~PINC)&0x07;
	int     dice;
	
	switch(rawDice)
	{
		case 1:
			dice = 20;
			break;
		case 2:
			dice = 12;
			break;
		case 3:
			dice = 10;
			break;
		case 4:
			dice = 8;
			break;
		case 5:
			dice = 6;
			break;
		case 6:
			dice = 4;
			break;
		case 7:
			dice = 2;
			break;
		case 8:
			dice = 99;
			break;
		default:
			dice = 100;
	}
	return(dice);
}

// Get the status of the roll button ******************************************
uint8_t getButton()
{
	uint8_t ret = ((~PINC)&0b10000000)>>7;
	return(ret);
}

// Program entry point ********************************************************
int main(void)
{
	unsigned int seed = 0, seeded = 0;
	int          i, count = 0, dice = 0, result = 0, timer = 0;
	State        state = State_Init;
	
	initIO();
	
    // Forever
    while (1) 
    {
		// Increment the seend counter
		++seed;
		
		// Implement the states of the application
		switch(state)
		{
			// Initialization
			case State_Init:
				count = getCount();
				dice = getDice();
				result = 0;
				timer = 0;
				state = State_Idle;
				break;
			// Idle
			case State_Idle:
				output(result);
				if(getDice() != dice)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Dice;
				}
				else if(getCount() != count)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Count;
				}
				else if(getButton())
				{
					timer = 10;
					state = State_Roll;
				}
				break;
			// Set the dice type to be rolled
			case State_Update_Dice:
				output(BLANK_OUTPUT);
				if(!--timer)
				{
					dice = getDice();
					timer = DISPLAY_CONFIG_TIME;
					result = 0;
					state = State_Display_Dice;
				}
				break;
			// Momentarily display the dice type
			case State_Display_Dice:
				if(dice == 100)
					output(99);
				else
					output(dice);
				if(getDice() != dice)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Dice;
				}
				else if(!--timer)
					state = State_Idle;
				break;
			// Set the number of dice to be rolled
			case State_Update_Count:
				output(BLANK_OUTPUT);
				if(!--timer)
				{
					count = getCount();
					timer = DISPLAY_CONFIG_TIME;
					result = 0;
					state = State_Display_Count;
				}
				break;
			// Momentarily display the dice count
			case State_Display_Count:
				output(count);
				if(getCount() != count)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Count;
				}
				else if(!--timer)
					state = State_Idle;
				break;
			// Simulate a roll by flashing random values of the dice type on the display
			case State_Roll:
				output((rand()%dice)+1);
				if(!--timer)
					state = State_Roll_Result;
				break;
			// Calculate and display the roll result
			case State_Roll_Result:
				// If not seeded...
				if(!seeded)
				{
					// Seed the rand() function with the seed counter
					// that has been incrementing since startup
					srand(seed);
					seeded = 1;
				}
				result = 0;
				for(i=0;i<count;++i)
					result+=(rand()%dice)+1;

				state = State_Idle;
				break;
			
		}
		// Insert delay based on the current state
		// Note: if init or idle state, no delay inserted and 
		// The seed counter increments as fast as this loop spins.
		if(state==State_Roll)
			delay_ms(100);
		else if(state>State_Idle)
			delay_ms(1);
    }
}

